{***********************************************************************************
                       PilotLogic Software House

FPC Package sqldb/firebird
This unit is part of CodeTyphon Studio (http://www.pilotlogic.com/)
************************************************************************************
}

unit FirebirdMon;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils,
  firebirddyn,
  firebirdConn,
  syncobjs,db,dbconst;

type
  TEventAlert = procedure(Sender: TObject; EventName: string; EventCount: longint;
    var CancelAlerts: boolean) of object;
  TErrorEvent = procedure(Sender: TObject; ErrorCode: integer) of object;

  { TFirebirdEventMonitor }

  TFirebirdEventMonitor=class (TComponent)
  private
    FConnection: TFirebirdConnection;
    FErrorMsg: string;
    FEvents: TStrings;
    FEventsThread: TThread;
    FOnError: TErrorEvent;
    FOnEventAlert: TEventAlert;
    FRegistered: Boolean;
    function GetNativeHandle: PISC_DB_HANDLE;
    procedure SetConnection(AValue: TFirebirdConnection);
    procedure SetEvents(AValue: TStrings);
    procedure SetRegistered(AValue: Boolean);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure RegisterEvents; virtual;
    procedure UnRegisterEvents; virtual;
    property ErrorMsg:string read FErrorMsg;
    property NativeHandle: PISC_DB_HANDLE read GetNativeHandle;
  published
    property Connection: TFirebirdConnection read FConnection write SetConnection;
    property Events: TStrings read FEvents write SetEvents;
    property Registered: Boolean read FRegistered write SetRegistered;
    property OnEventAlert: TEventAlert read FOnEventAlert write FOnEventAlert;
    property OnError: TErrorEvent read FOnError write FOnError;
  end;

implementation
const
  MAXEVENTSPEREPB=15;      //isc_event_block limitated to 15 events.
type
  TEPBpair=record
    EventBuf:PChar;        //isc_event_block event block
    ResultBuf:PChar;       //isc_event_block result block
    Signal:pointer;        //pointer to TFBEventsThread.FSignal
    Signaled:Boolean;      //this event block is signaled
    Len:ISC_LONG;          //lenght returned by isc_event_block
    Count:integer;         //number of events in this event block
    EventID:ISC_LONG;      //handle of current event returned by isc_que_events
    bStartup:boolean;      //results are not valid yet, don't generate event
  end;
  PEPBpair=^TEPBpair;

  { TFirebirdEventsThread }

  TFirebirdEventsThread=class(TThread)
  protected
    FCancelAlerts: Boolean;
    FCounts: array [0..19] of longint; //FB manual says ISC_STATUS but FB code int32
    FEPBs: array of TEPBpair;
    FErrorCode: integer;
    FEventCount: integer;
    FEventNumber: integer;
    FParent:TFirebirdEventMonitor;
    FSignal:TSimpleEvent;
    FStatus: array [0..19] of ISC_STATUS;
    procedure CheckError(Status: PISC_STATUS);
    procedure DoErrorEvent;
    procedure DoEventAlert;
    procedure Execute; override;
    procedure QueueEvents(DBHandle: pointer; block: integer);
  public
    constructor Create(Parent:TFirebirdEventMonitor);
    procedure DoTerminate;
  end;

  { THandleFirebirdConnection }
  // descendant to access protected GetHandle
  THandleFirebirdConnection=class(TFirebirdConnection)
  private
    function GetDBHandle: pointer;
  public
    property Handle:pointer read GetDBHandle;
  end;

procedure event_function(ptr:pointer;len:ushort;updated: pchar);cdecl;
begin
  Move(updated^, PEPBpair(ptr)^.ResultBuf^, len);
  PEPBpair(ptr)^.signaled:=true;
  TEventObject(PEPBpair(ptr)^.signal^).SetEvent;
end;

function THandleFirebirdConnection.GetDBHandle: pointer;
begin
  result:=GetHandle;
end;

{ TFirebirdEventsThread }

procedure TFirebirdEventsThread.CheckError(Status: PISC_STATUS);
var
  buf : array [0..1023] of char;

begin
  if ((Status[0] = 1) and (Status[1] <> 0)) then
  begin
    FErrorCode := Status[1];
    FParent.FErrorMsg := '';
    while isc_interprete(Buf, @Status) > 0 do
      FParent.FErrorMsg := FParent.FErrorMsg + LineEnding +' -' + StrPas(Buf);
    if Assigned(FParent.FOnError) then
      synchronize(@DoErrorEvent);
  end;
end;

procedure TFirebirdEventsThread.DoErrorEvent;
begin
  FParent.FOnError(FParent,FErrorCode);
end;

procedure TFirebirdEventsThread.DoEventAlert;
begin
  FParent.FOnEventAlert(FParent, FParent.FEvents[FEventNumber],FEventCount, FCancelAlerts);
end;

procedure TFirebirdEventsThread.Execute;
var
  i,j:integer;
  DBHandle : pointer;
  bStartup:boolean;

  function P(num:integer):pchar;
  begin
    num:=num+i*MAXEVENTSPEREPB;
    if num<FParent.FEvents.Count then
      result:=pchar(FParent.FEvents[num])
    else
      result:=nil;
  end;

begin
  // create event blocks and register events
  DBHandle:=THandleFirebirdConnection(FParent.Connection).GetDBHandle;
  SetLength(FEPBs,1+(FParent.FEvents.Count -1) div MAXEVENTSPEREPB);
  for i:=0 to high(FEPBs) do
    begin
    FEPBs[i].Signal:=@FSignal;
    FEPBs[i].bStartup:=true;
    FEPBs[i].Count:=FParent.FEvents.Count-i*MAXEVENTSPEREPB;
    if FEPBs[i].Count > MAXEVENTSPEREPB then
      FEPBs[i].Count:=MAXEVENTSPEREPB;
    FEPBs[i].Len := isc_event_block(@FEPBs[i].EventBuf, @FEPBs[i].ResultBuf,
        word(FEPBs[i].Count),[
        P(0), P(1),P(2), P(3), P(4),P(5), P(6), P(7),
        P(8), P(9), P(10), P(11), P(12), P(13), P(14)]);
    QueueEvents(DBHandle,i);
    end;
  FCancelAlerts:=false;
  bStartup:=true;
  while not Terminated do
    begin
    if FSignal.WaitFor(100)=wrSignaled then
      begin
      FSignal.ResetEvent;
      for i:=0 to high(FEPBs) do
        if FEPBs[i].Signaled then
          begin
          FEPBs[i].Signaled:=false;
          isc_event_counts(@FCounts[0], Short(FEPBs[i].Len), FEPBs[i].EventBuf, FEPBs[i].ResultBuf);
          if not FEPBs[i].bStartup and Assigned(FParent.FOnEventAlert) then
            begin
            for j:=0 to FEPBs[i].Count-1 do
              if (FCounts[j]<>0) and not FCancelAlerts then
                begin
                FEventNumber:=(i-low(FEPBs))*MAXEVENTSPEREPB+j;
                FEventCount:=FCounts[j];
                Synchronize(@DoEventAlert);
                end;
            end;
          FEPBs[i].bStartup:=false;
          QueueEvents(DBHandle,i);
          end;
      end;
    if terminated or FCancelAlerts or not FParent.Connection.Connected then
      break;
  end;
  // unregister events
  if FParent.Connection.Connected then   // Don't do this if connection is gone
    for i:=0 to high(FEPBs) do
      begin
      isc_cancel_events(@FStatus[0],@DBHandle,@FEPBs[i].EventID);
      isc_free(FEPBs[i].EventBuf);   //points to unreachable memory if connection gone
      isc_free(FEPBs[i].ResultBuf);
      CheckError(FStatus);
      end;
  SetLength(FEPBs,0);
  FSignal.destroy;
  FParent.FRegistered :=false;
end;

procedure TFirebirdEventsThread.QueueEvents(DBHandle:pointer;block: integer);
begin
  isc_que_events(@FStatus[0], @DBHandle, @FEPBs[block].EventID, Short(FEPBs[block].Len),
    FEPBs[block].EventBuf,TISC_CALLBACK(@event_function), @FEPBs[block]);
  CheckError(FStatus);
end;

constructor TFirebirdEventsThread.Create(Parent: TFirebirdEventMonitor);
begin
  FParent:=Parent;
  FSignal:=TSimpleEvent.Create();
  inherited create(false);
end;

procedure TFirebirdEventsThread.DoTerminate;
begin
  if not Terminated then
    begin
    Terminate;
    FSignal.SetEvent;
    end;
end;

{ TFirebirdEventMonitor }


function TFirebirdEventMonitor.GetNativeHandle: PISC_DB_HANDLE;
begin
  result:=THandleFirebirdConnection(FConnection).GetDBHandle;
end;


procedure TFirebirdEventMonitor.SetConnection(AValue: TFirebirdConnection);
begin
  if FConnection=AValue then Exit;
  If not (csDesigning in ComponentState) and FRegistered then
    begin
    if assigned(FConnection) then
      FConnection.RemoveFreeNotification(self); // remove us from the old connection
    UnRegisterEvents;
    FConnection:=AValue;
    if assigned(FConnection) then
      begin
      RegisterEvents;
      end;
    end
  else
    FConnection:=AValue;
  if assigned(FConnection) then
    FConnection.FreeNotification(Self); //in case Connection is destroyed before we are

end;

procedure TFirebirdEventMonitor.SetEvents(AValue: TStrings);
begin
  FEvents.Assign(AValue);
end;

procedure TFirebirdEventMonitor.SetRegistered(AValue: Boolean);
begin
  FRegistered := AValue;
  if not (csDesigning in ComponentState) then
    if AValue then
      RegisterEvents
    else
      UnRegisterEvents;
end;

constructor TFirebirdEventMonitor.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FEvents:=TStringList.Create;
  InitialiseFirebird;             // stick to library in case connection closes before us
end;

destructor TFirebirdEventMonitor.Destroy;
begin
  if FRegistered then
    UnRegisterEvents;
  if assigned(FConnection) then
    FConnection.RemoveFreeNotification(self);
  FEvents.Free;
  ReleaseFirebird;
  inherited Destroy;
end;

procedure TFirebirdEventMonitor.RegisterEvents;
begin
  If not assigned(FConnection) then
    DatabaseError(SErrNoDatabaseAvailable,Self);
  if not(csDesigning in ComponentState) and not FRegistered and (Events.Count>0) then
    begin
    if not Connection.Connected then
       Connection.Connected:=true;
    if Connection.Connected then
      begin
      FEventsThread:=TFirebirdEventsThread.Create(Self);
      FEventsThread.FreeOnTerminate:=true;
      FRegistered :=assigned(FEventsThread);
      end;
    end;
end;

procedure TFirebirdEventMonitor.UnRegisterEvents;
begin
  if not (csDesigning in ComponentState) and FRegistered then
    begin
    TFirebirdEventsThread(FEventsThread).DoTerminate;
    FEventsThread.WaitFor;
    FRegistered :=false;
    end;
end;

end.

